@@ -2,11 +2,22 @@ use clap::ArgMatches;
 use std::path::{Component, Path, PathBuf};
 
 use cargo::core::Workspace;
-use cargo_util::paths::{copy, create_dir_all};
-use semver::Version;
+use cargo_util::paths::{self, create_dir_all};
 
 use crate::build::*;
 use crate::build_targets::BuildTargets;
+use crate::target::Target;
+
+pub fn copy<P: AsRef<Path>, Q: AsRef<Path>>(ws: &Workspace, from: P, to: Q) -> anyhow::Result<u64> {
+    ws.gctx().shell().verbose(|shell| {
+        shell.status(
+            "Copying",
+            format!("{} to {}", from.as_ref().display(), to.as_ref().display()),
+        )
+    })?;
+
+    paths::copy(from, to)
+}
 
 fn append_to_destdir(destdir: Option<&Path>, path: &Path) -> PathBuf {
     if let Some(destdir) = destdir {
@@ -91,8 +102,9 @@ impl LibType {
             | ("android", _)
             | ("haiku", _)
             | ("illumos", _)
+            | ("openbsd", _)
             | ("emscripten", _) => LibType::So,
-            ("macos", _) | ("ios", _) => LibType::Dylib,
+            ("macos", _) | ("ios", _) | ("tvos", _) | ("visionos", _) => LibType::Dylib,
             ("windows", _) => LibType::Windows,
             _ => unimplemented!("The target {}-{} is not supported yet", os, env),
         }
@@ -101,43 +113,42 @@ impl LibType {
 
 pub(crate) struct UnixLibNames {
     canonical: String,
-    with_major_ver: String,
+    with_main_ver: String,
     with_full_ver: String,
 }
 
 impl UnixLibNames {
-    pub(crate) fn new(lib_type: LibType, lib_name: &str, lib_version: &Version) -> Option<Self> {
+    pub(crate) fn new(lib_type: LibType, library: &LibraryCApiConfig) -> Option<Self> {
+        let lib_name = &library.name;
+        let lib_version = &library.version;
+        let main_version = library.sover();
+
         match lib_type {
             LibType::So => {
                 let lib = format!("lib{lib_name}.so");
-                let lib_with_major_ver = format!("{}.{}", lib, lib_version.major);
                 let lib_with_full_ver = format!(
-                    "{}.{}.{}",
-                    lib_with_major_ver, lib_version.minor, lib_version.patch
+                    "{}.{}.{}.{}",
+                    lib, lib_version.major, lib_version.minor, lib_version.patch
                 );
+                let lib_with_main_ver = format!("{}.{}", lib, main_version);
+
                 Some(Self {
                     canonical: lib,
-                    with_major_ver: lib_with_major_ver,
+                    with_main_ver: lib_with_main_ver,
                     with_full_ver: lib_with_full_ver,
                 })
             }
             LibType::Dylib => {
                 let lib = format!("lib{lib_name}.dylib");
-                let lib_with_major_ver = if lib_version.major == 0 {
-                    format!(
-                        "lib{}.{}.{}.dylib",
-                        lib_name, lib_version.major, lib_version.minor
-                    )
-                } else {
-                    format!("lib{}.{}.dylib", lib_name, lib_version.major)
-                };
+                let lib_with_main_ver = format!("lib{}.{}.dylib", lib_name, main_version);
+
                 let lib_with_full_ver = format!(
                     "lib{}.{}.{}.{}.dylib",
                     lib_name, lib_version.major, lib_version.minor, lib_version.patch
                 );
                 Some(Self {
                     canonical: lib,
-                    with_major_ver: lib_with_major_ver,
+                    with_main_ver: lib_with_main_ver,
                     with_full_ver: lib_with_full_ver,
                 })
             }
@@ -146,12 +157,14 @@ impl UnixLibNames {
     }
 
     fn links(&self, install_path_lib: &Path) {
-        let mut ln_sf = std::process::Command::new("ln");
-        ln_sf.arg("-sf");
-        ln_sf
-            .arg(&self.with_full_ver)
-            .arg(install_path_lib.join(&self.with_major_ver));
-        let _ = ln_sf.status().unwrap();
+        if self.with_main_ver != self.with_full_ver {
+            let mut ln_sf = std::process::Command::new("ln");
+            ln_sf.arg("-sf");
+            ln_sf
+                .arg(&self.with_full_ver)
+                .arg(install_path_lib.join(&self.with_main_ver));
+            let _ = ln_sf.status().unwrap();
+        }
 
         let mut ln_sf = std::process::Command::new("ln");
         ln_sf.arg("-sf");
@@ -163,15 +176,16 @@ impl UnixLibNames {
 
     pub(crate) fn install(
         &self,
+        ws: &Workspace,
         capi_config: &CApiConfig,
         shared_lib: &Path,
         install_path_lib: &Path,
     ) -> anyhow::Result<()> {
         if capi_config.library.versioning {
-            copy(shared_lib, install_path_lib.join(&self.with_full_ver))?;
+            copy(ws, shared_lib, install_path_lib.join(&self.with_full_ver))?;
             self.links(install_path_lib);
         } else {
-            copy(shared_lib, install_path_lib.join(&self.canonical))?;
+            copy(ws, shared_lib, install_path_lib.join(&self.canonical))?;
         }
         Ok(())
     }
@@ -190,6 +204,7 @@ pub fn cinstall(ws: &Workspace, packages: &[CPackage]) -> anyhow::Result<()> {
             install_path_lib.push(subdir);
         }
 
+        let install_path_bin = append_to_destdir(destdir.as_deref(), &paths.bindir);
         let install_path_lib = append_to_destdir(destdir.as_deref(), &install_path_lib);
         let install_path_pc = append_to_destdir(destdir.as_deref(), &paths.pkgconfigdir);
         let install_path_include = append_to_destdir(destdir.as_deref(), &paths.includedir);
@@ -198,75 +213,93 @@ pub fn cinstall(ws: &Workspace, packages: &[CPackage]) -> anyhow::Result<()> {
         create_dir_all(&install_path_lib)?;
         create_dir_all(&install_path_pc)?;
 
-        ws.config()
-            .shell()
-            .status("Installing", "pkg-config file")?;
+        ws.gctx().shell().status("Installing", "pkg-config file")?;
 
         copy(
+            ws,
             &build_targets.pc,
             install_path_pc.join(build_targets.pc.file_name().unwrap()),
         )?;
 
         if capi_config.header.enabled {
-            ws.config().shell().status("Installing", "header file")?;
+            ws.gctx().shell().status("Installing", "header file")?;
             for (from, to) in build_targets.extra.include.iter() {
                 let to = install_path_include.join(to);
                 create_dir_all(to.parent().unwrap())?;
-                copy(from, to)?;
+                copy(ws, from, to)?;
             }
         }
 
         if !build_targets.extra.data.is_empty() {
-            ws.config().shell().status("Installing", "data file")?;
+            ws.gctx().shell().status("Installing", "data file")?;
             for (from, to) in build_targets.extra.data.iter() {
                 let to = install_path_data.join(to);
                 create_dir_all(to.parent().unwrap())?;
-                copy(from, to)?;
+                copy(ws, from, to)?;
             }
         }
 
         if let Some(ref static_lib) = build_targets.static_lib {
-            ws.config().shell().status("Installing", "static library")?;
-            copy(
-                static_lib,
-                install_path_lib.join(static_lib.file_name().unwrap()),
-            )?;
+            ws.gctx().shell().status("Installing", "static library")?;
+            let file_name = build_targets.static_output_file_name().unwrap();
+
+            copy(ws, static_lib, install_path_lib.join(file_name))?;
         }
 
         if let Some(ref shared_lib) = build_targets.shared_lib {
-            ws.config().shell().status("Installing", "shared library")?;
+            ws.gctx().shell().status("Installing", "shared library")?;
 
-            let lib_name = &capi_config.library.name;
             let lib_type = LibType::from_build_targets(build_targets);
             match lib_type {
                 LibType::So | LibType::Dylib => {
-                    let lib = UnixLibNames::new(lib_type, lib_name, &capi_config.library.version)
-                        .unwrap();
-                    lib.install(capi_config, shared_lib, &install_path_lib)?;
+                    let lib = UnixLibNames::new(lib_type, &capi_config.library).unwrap();
+                    lib.install(ws, capi_config, shared_lib, &install_path_lib)?;
                 }
                 LibType::Windows => {
-                    let lib_name = shared_lib.file_name().unwrap();
+                    let lib_name = build_targets.shared_output_file_name().unwrap();
 
                     if capi_config.library.install_subdir.is_none() {
                         let install_path_bin = append_to_destdir(destdir.as_deref(), &paths.bindir);
                         create_dir_all(&install_path_bin)?;
 
-                        copy(shared_lib, install_path_bin.join(lib_name))?;
+                        copy(ws, shared_lib, install_path_bin.join(lib_name))?;
                     } else {
                         // We assume they are plugins, install them in the custom libdir path
-                        copy(shared_lib, install_path_lib.join(lib_name))?;
+                        copy(ws, shared_lib, install_path_lib.join(lib_name))?;
                     }
                     if capi_config.library.import_library {
                         let impl_lib = build_targets.impl_lib.as_ref().unwrap();
-                        let impl_lib_name = impl_lib.file_name().unwrap();
-                        copy(impl_lib, install_path_lib.join(impl_lib_name))?;
+                        let impl_lib_name = if build_targets.use_meson_naming_convention {
+                            format!("{}.lib", build_targets.name).into()
+                        } else {
+                            impl_lib.file_name().unwrap().to_owned()
+                        };
+                        copy(ws, impl_lib, install_path_lib.join(impl_lib_name))?;
                         let def = build_targets.def.as_ref().unwrap();
                         let def_name = def.file_name().unwrap();
-                        copy(def, install_path_lib.join(def_name))?;
+                        copy(ws, def, install_path_lib.join(def_name))?;
                     }
                 }
             }
         }
+
+        if let Some(ref debug_info) = build_targets.debug_info {
+            if debug_info.exists() {
+                ws.gctx()
+                    .shell()
+                    .status("Installing", "debugging information")?;
+                let destination_path = build_targets
+                    .debug_info_file_name(&install_path_bin, &install_path_lib)
+                    .unwrap();
+
+                create_dir_all(destination_path.parent().unwrap())?;
+                copy(ws, debug_info, destination_path)?;
+            } else {
+                ws.gctx()
+                    .shell()
+                    .verbose(|shell| shell.status("Absent", "debugging information"))?;
+            }
+        }
     }
 
     Ok(())
@@ -284,39 +317,46 @@ pub struct InstallPaths {
     pub pkgconfigdir: PathBuf,
 }
 
+fn get_path_or(args: &ArgMatches, id: &str, f: impl FnOnce() -> PathBuf) -> PathBuf {
+    if matches!(
+        args.value_source(id),
+        Some(clap::parser::ValueSource::DefaultValue)
+    ) {
+        f()
+    } else {
+        args.get_one::<PathBuf>(id).unwrap().to_owned()
+    }
+}
+
 impl InstallPaths {
-    pub fn new(_name: &str, args: &ArgMatches, capi_config: &CApiConfig) -> Self {
+    pub fn new(
+        _name: &str,
+        rustc_target: &Target,
+        args: &ArgMatches,
+        capi_config: &CApiConfig,
+    ) -> Self {
         let destdir = args.get_one::<PathBuf>("destdir").map(PathBuf::from);
-        let prefix = args
-            .get_one::<PathBuf>("prefix")
-            .map(PathBuf::from)
-            .unwrap_or_else(|| "/usr/local".into());
-        let libdir = args
-            .get_one::<PathBuf>("libdir")
-            .map(PathBuf::from)
-            .unwrap_or_else(|| prefix.join("lib"));
-        let includedir = args
-            .get_one::<PathBuf>("includedir")
-            .map(PathBuf::from)
-            .unwrap_or_else(|| prefix.join("include"));
-        let datarootdir = args
-            .get_one::<PathBuf>("datarootdir")
-            .map(PathBuf::from)
-            .unwrap_or_else(|| prefix.join("share"));
+        let prefix = get_path_or(args, "prefix", || rustc_target.default_prefix());
+        let libdir = prefix.join(get_path_or(args, "libdir", || {
+            rustc_target.default_libdir()
+        }));
+        let includedir = prefix.join(get_path_or(args, "includedir", || {
+            rustc_target.default_includedir()
+        }));
+        let datarootdir = prefix.join(get_path_or(args, "datarootdir", || {
+            rustc_target.default_datadir()
+        }));
         let datadir = args
             .get_one::<PathBuf>("datadir")
-            .map(PathBuf::from)
+            .map(|d| prefix.join(d))
             .unwrap_or_else(|| datarootdir.clone());
 
         let subdir_name = PathBuf::from(&capi_config.header.subdirectory);
 
-        let bindir = args
-            .get_one::<PathBuf>("bindir")
-            .map(PathBuf::from)
-            .unwrap_or_else(|| prefix.join("bin"));
+        let bindir = prefix.join(args.get_one::<PathBuf>("bindir").unwrap());
         let pkgconfigdir = args
             .get_one::<PathBuf>("pkgconfigdir")
-            .map(PathBuf::from)
+            .map(|d| prefix.join(d))
             .unwrap_or_else(|| libdir.join("pkgconfig"));
 
         InstallPaths {
